import collections
import copy

from jsbgym import rewards
from jsbgym.tasks import FlightTask
from jsbgym.rewards import State, Reward, RewardComponent
from jsbgym.assessors import Assessor
from jsbgym.simulation import Simulation
import jsbgym.properties as prp
from typing import Tuple, NamedTuple, Iterable, Dict


class AssessorStub(Assessor):
    def assess(self, state: State, prev_state: State, is_terminal: bool) -> Reward:
        base_reward = (0,)
        shaping_reward = ()
        return Reward(base_reward, shaping_reward)


class FlightTaskStub(FlightTask):
    """A minimal Task for testing."""

    test_property1 = prp.BoundedProperty(
        "test_property1", "dummy property for testing", -1, 1
    )
    test_property2 = prp.BoundedProperty(
        "test_property2", "dummy property for testing", -1, 1
    )

    def __init__(self, *_):
        self.state_variables = (self.test_property1, self.test_property2)
        self.action_variables = (prp.aileron_cmd, prp.elevator_cmd)
        super().__init__(AssessorStub())

    def _is_terminal(self, _: Tuple[float, ...], __: float) -> bool:
        return False

    def get_initial_conditions(self):
        return self.base_initial_conditions

    def get_state(self, value1: float, value2: float):
        """Returns a State of this class' test properties populated with input values"""
        return self.State(value1, value2)

    def get_props_to_output(self) -> Tuple:
        return prp.u_fps, prp.altitude_sl_ft, prp.heading_deg

    @staticmethod
    def get_dummy_state_class_and_properties(length: int):
        dummy_properties = tuple(
            prp.Property("test_prop" + str(i), "") for i in range(length)
        )
        DummyState = collections.namedtuple(
            "DummyState", [prop.name for prop in dummy_properties]
        )
        return DummyState, dummy_properties

    @staticmethod
    def get_dummy_state_and_properties(
        values: Iterable[float],
    ) -> Tuple[NamedTuple, Tuple[prp.Property, ...]]:
        """
        given a collection of floats, creates dummy Properties for each value
        and inits a State
        """
        values_safe = tuple(values)
        DummyState, props = FlightTaskStub.get_dummy_state_class_and_properties(
            len(values_safe)
        )
        return DummyState(*values_safe), props

    def _reward_terminal_override(self, reward: rewards.Reward, sim: Simulation):
        return False


class BasicFlightTask(FlightTask):
    """A Task with basic but realistic state and action space."""

    def __init__(self, *_):
        self.state_variables = super().base_state_variables
        self.action_variables = (prp.aileron_cmd, prp.rudder_cmd, prp.elevator_cmd)
        super().__init__(AssessorStub())

    def _is_terminal(self, _: Simulation) -> bool:
        return False

    def get_initial_conditions(self):
        return self.base_initial_conditions

    def get_props_to_output(self) -> Tuple:
        return prp.u_fps, prp.altitude_sl_ft, prp.heading_deg

    def _reward_terminal_override(
        self, reward: rewards.Reward, sim: Simulation
    ) -> bool:
        return False


class SimStub(object):
    def __init__(self):
        self.data = {}

    def run(self):
        pass

    def start_engines(self):
        self[prp.engine_running] = 1.0

    def raise_landing_gear(self):
        self[prp.gear] = 0.0
        self[prp.gear_all_cmd] = 0.0

    def get_sim_time(self) -> float:
        """Gets the simulation time from JSBSim, a float."""
        return self[prp.sim_time_s]

    def __setitem__(self, prop: prp.Property, value: float):
        return self.data.__setitem__(prop.name, value)

    def __getitem__(self, prop: prp.Property) -> float:
        return self.data.__getitem__(prop.name)

    def copy(self):
        new = SimStub()
        new.data = copy.deepcopy(self.data)
        return new

    def set_throttle_mixture_controls(self, throttle_cmd, mixture_cmd):
        self[prp.throttle_cmd] = throttle_cmd
        self[prp.mixture_cmd] = mixture_cmd

        self[prp.throttle_1_cmd] = throttle_cmd
        self[prp.mixture_1_cmd] = mixture_cmd

    @staticmethod
    def make_valid_state_stub(task: FlightTask):
        """
        A factory method for SimStubs with valid state values for a task.

        Returns a SimStub containing keys of the names of state properties
        mapped to valid values at the midpoint of the allowable range of
        each property.

        :param task: a TypeModule, the task the SimStub should conform to
        :return: a SimStub configured with valid state for the task
        """
        sim = SimStub()
        for initial_prop, value in task.get_initial_conditions().items():
            sim[initial_prop] = value
        for prop in task.state_variables:
            typical_value = (prop.min + prop.max) / 2
            sim[prop] = typical_value
        sim[prp.sim_time_s] = 1.0
        sim[prp.lat_geod_deg] = task.get_initial_conditions()[
            prp.initial_latitude_geod_deg
        ]
        sim[prp.lng_geoc_deg] = task.get_initial_conditions()[
            prp.initial_longitude_geoc_deg
        ]
        sim[prp.dist_travel_m] = 2.0
        sim[prp.heading_deg] = 270
        sim[prp.v_north_fps] = 0
        sim[prp.v_east_fps] = -200.0  # corresp. to travel at track 270 deg
        sim[prp.altitude_sl_ft] = task.INITIAL_ALTITUDE_FT
        return sim


class TransitioningSimStub(object):
    """
    A dummy Simulation object which holds two SimStub states, which it
    transitions between when run() or reset().
    """

    def __init__(self, initial_sim: SimStub, next_sim: SimStub):
        self.initial_sim = initial_sim
        self.next_sim = next_sim
        self.current_sim = self.initial_sim

    def reset(self):
        self.current_sim = self.initial_sim

    def run(self):
        self.current_sim = self.next_sim

    def __setitem__(self, prop: prp.Property, value: float):
        self.current_sim[prop] = value

    def __getitem__(self, prop: prp.Property) -> float:
        return self.current_sim[prop]

    def start_engines(self):
        self.current_sim[prp.engine_running] = 1.0


class DefaultSimStub(object):
    """
    A stub for a Sim object which never throws KeyErrors when retrieving
    properties; a default value is always returned instead.
    """

    def __init__(self, default_value=5.0):
        self.default_value = default_value
        self.properties = {}

    def __getitem__(self, prop):
        return self.properties.get(prop.name, self.default_value)

    def __setitem__(self, prop, value):
        self.properties[prop.name] = value

    def run(self):
        pass

    def start_engines(self):
        pass


class ConstantRewardComponentStub(RewardComponent):
    """A RewardComponent which always returns a constant value"""

    default_return = 1.0

    def __init__(self, return_value: float = default_return):
        self.return_value = return_value

    def calculate(self, _: State, __: State, ___: bool):
        return self.return_value

    def get_potential(self, state: State, is_terminal):
        return self.return_value

    def is_potential_difference_based(self):
        return False

    def get_return_value(self):
        return self.return_value

    def get_name(self):
        return str(self)


class RewardComponentStub(RewardComponent):
    """
    A stub PotentialBasedComponent which is preconfigured to return a specified
    potential for a given State
    """

    def __init__(self, state_potentials: Dict[State, float]):
        """
        :param state_potentials: dict mapping States to their potentials
        """
        self.state_potentials = state_potentials

    def calculate(self, state: State, last_state: State, is_terminal: bool):
        potential = self.get_potential(state, is_terminal)
        last_potential = self.get_potential(last_state, False)
        return potential - last_potential

    def get_potential(self, state: State, is_terminal):
        if is_terminal:
            return 0
        else:
            return self.state_potentials[state]

    def get_name(self):
        return str(self)

    def is_potential_difference_based(self) -> bool:
        return True



import unittest
import time
import jsbgym.properties as prp
from typing import Type
from jsbgym import tasks, aircraft
from jsbgym.tests.stubs import BasicFlightTask
from jsbgym.environment import JsbSimEnv
from jsbgym.agents import RandomAgent, ConstantAgent


class TestJsbSimInstance(unittest.TestCase):
    def setUp(self, task_type: Type[tasks.HeadingControlTask] = BasicFlightTask):
        self.env = None
        self.env = JsbSimEnv(task_type)
        self.env.reset()

    def tearDown(self):
        self.env.close()

    def test_long_episode_random_actions(self):
        self.setUp()
        tic = time.time()
        self.env.reset()
        for i in range(2000):
            self.env.step(action=self.env.action_space.sample())
            print(f"jsbsim {i / 10} s\n")
        toc = time.time()
        wall_time = toc - tic
        sim_time = self.env.sim.get_sim_time()
        print(f"Simulated {sim_time} s of flight in {wall_time} s")

    def test_render_episode(self):
        self.setUp()
        render_every = 5
        self.env.reset()
        for i in range(1000):
            action = self.env.action_space.sample()
            obs, _, _, _ = self.env.step(action=action)
            if i % render_every == 0:
                self.env.render()

    def test_render_steady_level_flight_random(self):
        """Runs steady level flight task with a random agent."""
        self.setUp(task_type=tasks.HeadingControlTask)
        agent = RandomAgent(self.env.action_space)
        render_every = 5
        ep_reward = 0
        terminated = False
        state = self.env.reset()
        step_number = 0
        while not terminated:
            action = agent.act(state)
            state, reward, temrinated, truncated, info = self.env.step(action)
            ep_reward += reward
            if step_number % render_every == 0:
                self.env.render()
            step_number += 1

    def test_run_episode_steady_level_flight_no_render(self):
        self.setUp(task_type=tasks.HeadingControlTask)
        agent = RandomAgent(self.env.action_space)
        report_every = 20
        EPISODES = 10

        for _ in range(EPISODES):
            ep_reward = 0
            terminated = False
            state = self.env.reset()
            step_number = 0
            while not terminated:
                action = agent.act(state)
                state, reward, terminated, info = self.env.step(action)
                ep_reward += reward
                if step_number % report_every == 0:
                    print(f"time:\t{self.env.sim.get_sim_time()} s")
                    print(f"last reward:\t{reward}")
                    print(f"episode reward:\t{ep_reward}")
                step_number += 1


class FlightGearRenderTest(unittest.TestCase):
    def setUp(
        self,
        plane: aircraft.Aircraft = aircraft.cessna172P,
        task_type: Type[tasks.HeadingControlTask] = tasks.HeadingControlTask,
    ):
        self.env = None
        self.env = JsbSimEnv(aircraft=plane, task_type=task_type)
        self.env.reset()

    def tearDown(self):
        self.env.close()

    def test_render_steady_level_flight(self):
        self.setUp(plane=aircraft.cessna172P, task_type=tasks.HeadingControlTask)
        agent = ConstantAgent(self.env.action_space)
        render_every = 5
        report_every = 20
        EPISODES = 999

        for _ in range(EPISODES):
            ep_reward = 0
            terminated = False
            state = self.env.reset()
            self.env.render(mode="flightgear")
            step_number = 0
            while not terminated:
                action = agent.act(state)
                state, reward, terminated, info = self.env.step(action)
                ep_reward += reward
                if step_number % render_every == 0:
                    self.env.render(mode="flightgear")
                if step_number % report_every == 0:
                    print(f"time:\t{self.env.sim.get_sim_time()} s")
                    print(f"last reward:\t{reward}")
                    print(f"episode reward:\t{ep_reward}")
                    print(f"thrust:\t{self.env.sim[prp.engine_thrust_lbs]}")
                    print(f"engine running:\t{self.env.sim[prp.engine_running]}")
                step_number += 1
            print(f"***\n EPISODE REWARD: {ep_reward}\n ***")


class TurnHeadingControlTest(unittest.TestCase):
    def setUp(
        self,
        plane: aircraft.Aircraft = aircraft.cessna172P,
        task_type: Type[tasks.HeadingControlTask] = tasks.TurnHeadingControlTask,
        shaping: tasks.Shaping = tasks.Shaping.STANDARD,
    ):
        self.env = None
        self.env = JsbSimEnv(aircraft=plane, task_type=task_type, shaping=shaping)
        self.env.reset()

    def tearDown(self):
        self.env.close()

    def test_render_heading_control(self):
        self.setUp(
            plane=aircraft.a320,
            task_type=tasks.TurnHeadingControlTask,
            shaping=tasks.Shaping.EXTRA_SEQUENTIAL,
        )
        agent = RandomAgent(self.env.action_space)
        render_every = 5
        report_every = 20
        EPISODES = 50

        for _ in range(EPISODES):
            ep_reward = 0
            terminated = False
            state = self.env.reset()
            self.env.render(mode="flightgear")
            step_number = 0
            while not terminated:
                action = agent.act(state)
                state, reward, terminated, truncated, info = self.env.step(action)
                ep_reward += reward
                if step_number % render_every == 0:
                    self.env.render(mode="flightgear")
                if step_number % report_every == 0:
                    heading_target = tasks.HeadingControlTask.target_track_deg
                    print(f"time:\t{self.env.sim.get_sim_time()} s")
                    print(f"last reward:\t{reward}")
                    print(f"episode reward:\t{ep_reward}")
                    print(f"gear status:\t{self.env.sim[prp.gear]}")
                    print(f"thrust eng0:\t{self.env.sim[prp.engine_thrust_lbs]}")
                    print(
                        f'thrust eng1:\t {self.env.sim[prp.Property("propulsion/engine[1]/thrust-lbs", "")]}'
                    )
                    print(f"heading:\t{self.env.sim[prp.heading_deg]}")
                    print(f"target heading:\t{self.env.sim[heading_target]}")
                    print("\n")
                step_number += 1
            print(f"***\n EPISODE REWARD: {ep_reward}\n ***\n")



import unittest
import gymnasium as gym
import numpy as np
import matplotlib.pyplot as plt
import jsbgym.properties as prp
import jsbgym
from jsbgym import aircraft, tasks, utils
from jsbgym.environment import JsbSimEnv, NoFGJsbSimEnv
from jsbgym.tests.stubs import BasicFlightTask
from jsbgym.visualiser import FlightGearVisualiser


class TestJsbSimEnv(unittest.TestCase):
    def setUp(self, agent_interaction_freq: int = 10):
        gym.logger.set_level(gym.logger.DEBUG)
        self.env = None
        self.init_env(agent_interaction_freq)
        self.env.reset()

    def init_env(self, agent_interaction_freq):
        self.env = JsbSimEnv(
            task_type=BasicFlightTask, agent_interaction_freq=agent_interaction_freq
        )

    def tearDown(self):
        self.env.close()

    def assertValidObservation(self, obs: np.array):
        """Helper; checks shape and values of an observation."""
        self.assertEqual(
            self.env.observation_space.shape,
            obs.shape,
            msg="observation has wrong size",
        )
        self.assert_in_box_space(obs, self.env.observation_space)

    def assertValidAction(self, action: np.array):
        """Helper; checks shape and values of an action."""
        self.assertEqual(
            self.env.action_space.shape, action.shape, msg="action has wrong size"
        )
        self.assert_in_box_space(action, self.env.action_space)

    def assert_in_box_space(self, sample: np.array, space: gym.spaces.Box) -> None:
        if space.contains(sample):
            return
        else:
            is_too_low = sample < space.low
            is_too_high = sample > space.high
            msg = "Sample is not in space:"
            for i in range(len(sample)):
                if is_too_low[i]:
                    msg += f"\nelement {i} too low: {sample[i]} < {space.low[i]}"
                if is_too_high[i]:
                    msg += f"\nelement {i} too high: {sample[i]} > {space.high[i]}"
            raise AssertionError(msg)

    def validate_action_made(self, action: np.array):
        """Helper; confirms action was correctly input to simulation."""
        self.assertValidAction(action)
        for prop, command in zip(self.env.task.action_variables, action):
            actual = self.env.sim[prop]
            self.assertAlmostEqual(
                command, actual, msg="simulation commanded value does not match action"
            )

    def test_init_spaces(self):
        # check correct types for obs and action space
        self.assertIsInstance(
            self.env.observation_space,
            gym.Space,
            msg="observation_space is not a Space object",
        )
        self.assertIsInstance(
            self.env.action_space, gym.Space, msg="action_space is not a Space object"
        )

        # check low and high values are as expected
        obs_lows = self.env.observation_space.low
        obs_highs = self.env.observation_space.high
        act_lows = self.env.action_space.low
        act_highs = self.env.action_space.high

        places_tol = 3

        for prop, lo, hi in zip(self.env.task.state_variables, obs_lows, obs_highs):
            self.assertAlmostEqual(
                lo,
                prop.min,
                msg=f"{prop} min of {prop.min} does not match space low of {lo}",
            )
            self.assertAlmostEqual(
                hi,
                prop.max,
                msg=f"{prop} max of {prop.max} does not match space high of {hi}",
            )

    def test_reset_env(self):
        self.setUp()
        obs = self.env.reset()

        self.assertValidObservation(obs)

    def test_do_action(self):
        self.setUp()
        action1 = np.array([0.0] * len(self.env.task.action_variables))
        action2 = np.linspace(-0.5, 0.5, num=len(self.env.task.action_variables))

        # do an action and check results
        obs, _, _, _ = self.env.step(action1)
        self.assertValidObservation(obs)
        self.validate_action_made(action1)

        # repeat action several times
        for _ in range(10):
            obs, _, _, _ = self.env.step(action1)
            self.assertValidObservation(obs)
            self.validate_action_made(action1)

        # repeat new action
        for _ in range(10):
            obs, _, _, _ = self.env.step(action2)
            self.assertValidObservation(obs)
            self.validate_action_made(action2)

    def test_figure_created_closed(self):
        self.env.render()
        self.assertIsInstance(self.env.figure_visualiser.figure, plt.Figure)
        self.env.close()
        self.assertIsNone(self.env.figure_visualiser.figure)

    def test_plot_state(self):
        # note: this checks that plot works without throwing exception
        # correctness of plot must be checked in appropriate manual_test
        self.setUp()
        self.env.render()

        action = np.array([0.0] * len(self.env.task.action_variables))
        # repeat action several times
        for _ in range(3):
            obs, _, _, _ = self.env.step(action)
            self.env.render()

    def test_plot_actions(self):
        # note: this checks that plot works without throwing exception
        # correctness of plot must be checked in appropriate manual_test
        self.env.render()

        # repeat action several times
        for _ in range(3):
            action = self.env.action_space.sample()
            _, _, _, _ = self.env.step(action)
            self.env.render()

    def test_asl_agl_elevations_equal(self):
        # we want the height above sea level to equal ground elevation at all times
        self.setUp(agent_interaction_freq=1)
        for i in range(25):
            self.env.step(action=self.env.action_space.sample())
            alt_sl = self.env.sim[prp.altitude_sl_ft]
            alt_gl = self.env.sim[prp.BoundedProperty("position/h-agl-ft", "", 0, 0)]
            self.assertAlmostEqual(alt_sl, alt_gl)

    def test_render_flightgear_mode(self):
        self.setUp()
        self.env.render(flightgear_blocking=False)
        self.assertIsInstance(self.env.flightgear_visualiser, FlightGearVisualiser)
        self.env.close()


class TestNoFlightGearJsbSimEnv(TestJsbSimEnv):
    def init_env(self, agent_interaction_freq):
        self.env = NoFGJsbSimEnv(
            task_type=BasicFlightTask, agent_interaction_freq=agent_interaction_freq
        )

    def test_render_flightgear_mode(self):
        with self.assertRaises(ValueError):
            self.env.render()


class TestGymRegistration(unittest.TestCase):
    def get_number_of_envs(self) -> int:
        num_tasks = 2
        num_flightgear_setttings = 2
        num_shaping_settings = 3
        num_aircraft = 3

        return (
            num_tasks * num_flightgear_setttings * num_shaping_settings * num_aircraft
        )

    def test_expected_number_of_envs_from_helper_function(self):
        expected_envs = self.get_number_of_envs()
        num_envs = len(utils.get_env_id_kwargs_map())

        self.assertLessEqual(expected_envs, num_envs)

    def test_expected_number_of_envs_from_enum(self):
        expected_envs = self.get_number_of_envs()
        num_envs = len(jsbgym.Envs)

        self.assertLessEqual(expected_envs, num_envs)

    def test_gym_environments_makeable_by_gym_from_helper_function(self):
        for jsb_env_id in utils.get_env_id_kwargs_map():
            env = gym.make(jsb_env_id)
            self.assertIsInstance(env, JsbSimEnv)

    def test_gym_environment_makeable_by_gym_from_enum(self):
        for jsb_env_id in jsbgym.Envs:
            env = gym.make(jsb_env_id.value)
            self.assertIsInstance(env, JsbSimEnv)

    def test_gym_environments_configured_correctly(self):
        Shaping = tasks.Shaping
        for task in (tasks.HeadingControlTask, tasks.TurnHeadingControlTask):
            for plane in (aircraft.cessna172P, aircraft.f15, aircraft.a320):
                for shaping in (Shaping.STANDARD, Shaping.EXTRA_SEQUENTIAL):
                    for enable_flightgear in (True, False):
                        id = utils.get_env_id(task, plane, shaping, enable_flightgear)
                        env = gym.make(id)

                        if enable_flightgear:
                            self.assertIsInstance(env, JsbSimEnv)
                        else:
                            self.assertIsInstance(env, NoFGJsbSimEnv)
                        self.assertIsInstance(env.task, task)
                        self.assertEqual(env.aircraft, plane)
